/*
 * Copyright (C) 2017 Du-Lab Team <dulab.binf@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package dulab.adap.workflow.decomposition;

import dulab.adap.datamodel.BetterComponent;
import dulab.adap.datamodel.BetterPeak;
import dulab.adap.datamodel.Chromatogram;
import dulab.adap.datamodel.Spectrum;
import org.dulab.javanmf.algorithms.MatrixFactorization;
import org.dulab.javanmf.algorithms.MatrixRegression;
import org.dulab.javanmf.updaterules.MUpdateRule;
import org.dulab.javanmf.updaterules.UpdateRule;
import org.jblas.DoubleMatrix;
import org.jblas.MatrixFunctions;

import javax.annotation.Nonnull;
import java.util.*;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * This class finds components (bi-gaussians) that best describe given peaks.
 * <p>
 * Step 1. Adjust the apex retention time of each peak by fitting a parabola to the top points of the peak
 * and calculating the center of that parabola.
 * <p>
 * Step 2. Cluster the adjusted apex retention times, using the hierarchical clustering with the ward-linkage.
 * <p>
 * Step 3. For each cluster, construct a bi-gaussian as an "average" of all peaks in a cluster.
 * <p>
 * Step 4. Adjust standard deviations of the bi-gaussians by minimizing the cost function ||X - W(s) x H||,
 * where X represents the detected peaks, W represents bi-gaussians with standard deviations s,
 * H represents the decomposition coefficients. The optimization is performed by alternating minimization
 * of two problems:
 * 4.1. Non-linear least squares problem argmin ||X - W(s) x H||^2 with respect to s,
 * 4.2. Linear least squares problem argmin ||X - W(s) x H||^2 with respect to H.
 * <p>
 * Step 5. Decompose every chromatogram into a linear combination of constructed components (bi-gaussians).
 *
 * @author Du-Lab Team <dulab.binf@gmail.com>
 */
public class ComponentSelector {

    private static final Logger LOG = Logger.getLogger(ComponentSelector.class.getName());

    private static final double[] SMOOTHING_COEFFICIENTS = new double[]{0.2, 0.2, 0.2, 0.2, 0.2};

    private boolean cancel;

    private final MatrixFactorization factorization;
    private final MatrixRegression leastSquares;

    /**
     * Creates an instance of the class for finding component that best fit given real peaks.
     */
    public ComponentSelector() {
        this.cancel = false;

        // We calculate lambda as
        UpdateRule updateRuleH = new MUpdateRule(0.0, 0.0);
        UpdateRule updateRuleW = new MUpdateRule(0.0, 0.0) {
            @Override
            public double update(DoubleMatrix xt, DoubleMatrix ht, DoubleMatrix wt) {
                double v = super.update(xt, ht, wt);
                makeUnimodalRows(wt);
                flattenLastRow(wt);
                scaleRows(wt);
                return v;
            }
        };
        factorization = new MatrixFactorization(updateRuleW, updateRuleH, 1e-12, 40000);  // 1e-12

        // Create an instance of MatrixRegression
        UpdateRule updateRule = new MUpdateRule(0.0, 0.0);
        leastSquares = new MatrixRegression(updateRule, 1e-10, 40000);  // 1e-10
    }

    public void setCancel(boolean cancel) {
        this.cancel = cancel;
    }

    /**
     * Finds components, their elution profiles and spectra.
     * <p>
     * <ol>
     * <li>Use retention-time clustering to determine the number of componets and their approximate elution profiles</li>
     * <li>Perform the non-negative matrix factroization to construct elution profiles of the components</li>
     * <li>Perform decomposition of all chromatograms into the linear combination of components and the baseline</li>
     * <li>Create a list of {@link BetterComponent} for each component with its elution profile and spectrum</li>
     * </ol>
     *
     * @return list of {@link BetterComponent} containing components with their elution profiles and spectra
     */
    @Nonnull
    public List<BetterComponent> execute(List<BetterPeak> chromatograms,
                                         RetTimeClusterer.Cluster cluster,
                                         double retTimeTolerance,
                                         boolean adjustApexRetTime,
                                         int minClusterSize) {

        this.cancel = false;

        // BiGaussianDetector performs adjustment of the apex retention times, cluster those apex retention times
        // and construct a bi-gaussian for each cluster
        BiGaussianDetector detector = new BiGaussianDetector(chromatograms, cluster,
                retTimeTolerance, adjustApexRetTime, minClusterSize);

        // Sorted array of all retention times
        DoubleMatrix retTimes = new DoubleMatrix(detector.retTimes);

        // Matrix representing detected peaks
        DoubleMatrix peakMatrix = peaksToMatrix(detector.peaks, false);

        DoubleMatrix heights = peakMatrix.columnMaxs();
        peakMatrix.muliRowVector(MatrixFunctions.sqrt(heights).divi(heights));

        // Parameters of constructed bi-gaussians
        List<BiGaussian> biGaussians = detector.biGaussians;

        if (biGaussians.isEmpty())
            return new ArrayList<>(0);

        // Construct the matrix representing the components (bi-gaussians)
        DoubleMatrix componentMatrix = biGaussiansToMatrix(biGaussians, retTimes.toArray());
        DoubleMatrix extendedComponentMatrix = addColumn(componentMatrix, 1.0);

        DoubleMatrix coefficientMatrix = getCoefficientMaxs(peakMatrix, componentMatrix);
        DoubleMatrix extendedCoefficientMatrix = addRow(coefficientMatrix, peakMatrix.max() * 1e-12);

        factorization.execute(peakMatrix, extendedComponentMatrix, extendedCoefficientMatrix, true);

        copyMatrix(extendedComponentMatrix, componentMatrix);
        copyMatrix(extendedCoefficientMatrix, coefficientMatrix);

        // --------------------------------------------------
        // ----- Perform decomposition of chromatograms -----
        // --------------------------------------------------

        double startRetTime = cluster.start;
        double endRetTime = cluster.end;
        double[] allRetTimes = getRetTimes(chromatograms, startRetTime, endRetTime);

        List<BetterPeak> filteredChromatograms = chromatograms.stream()
                .filter(c -> IntStream.range(0, c.chromatogram.length)
                        .filter(i -> startRetTime <= c.chromatogram.xs[i])
                        .filter(i -> c.chromatogram.xs[i] <= endRetTime)
                        .mapToDouble(i -> c.chromatogram.ys[i])
                        .sum() > 0.0)
                .collect(Collectors.toList());

        // Construct the matrix representing all chromatograms
        DoubleMatrix allMatrix = peaksToMatrix(new PeakList(filteredChromatograms, allRetTimes, false));

        // Construct the matrix representing components
//        DoubleMatrix allComponentMatrix = biGaussiansToMatrix(biGaussians, allRetTimes);
        DoubleMatrix allComponentMatrix = componentMatrix;
        // Remove components identically equal to zero
        allComponentMatrix = allComponentMatrix.getColumns(allComponentMatrix.columnMaxs().ne(0.0));
        final int numComponents = allComponentMatrix.columns;

        // Scale components
        allComponentMatrix.diviRowVector(allComponentMatrix.columnMaxs());
        DoubleMatrix allSpectraMatrix = getCoefficientMaxs(allMatrix, allComponentMatrix);

        // Add the baseline to components
        DoubleMatrix componentsWithBaseline = addColumn(allComponentMatrix, 1.0);
        DoubleMatrix limit = addRow(allSpectraMatrix, Double.MAX_VALUE);
        DoubleMatrix spectra = addRow(allSpectraMatrix, 1e-12 * allSpectraMatrix.max());

        leastSquares.solve(allMatrix, componentsWithBaseline, spectra, limit, true);

        // Get all mzValues
        double[] mzValues = filteredChromatograms.stream()
                .mapToDouble(BetterPeak::getMZ)
                .toArray();

        // Create a list of {@link BetterComponent}
        List<BetterComponent> componentsList = new ArrayList<>(numComponents);
        for (int i = 0; i < numComponents; ++i) {

            double maxHeight = Math.max(0.0, spectra.getRow(i).max());

            DoubleMatrix component = allComponentMatrix.getColumn(i);

            double[] componentsRetTimes = IntStream.range(0, component.length)
                    .filter(j -> component.get(j) > 0.0)
                    .mapToDouble(j -> allRetTimes[j])
                    .toArray();

            double[] componentsIntensities = IntStream.range(0, component.length)
                    .mapToDouble(component::get)
                    .filter(intensity -> intensity > 0.0)
                    .toArray();

            Chromatogram chromatogram = new Chromatogram(componentsRetTimes, componentsIntensities);
            chromatogram.scale(maxHeight / component.max());
            DoubleMatrix spectrum = spectra.getRow(i);

            componentsList.add(new BetterComponent(i, chromatogram,
                    new Spectrum(mzValues, spectrum.toArray()),
                    filteredChromatograms.get(spectrum.argmax()).info));
        }

        return componentsList;
    }

    private DoubleMatrix peaksToMatrix(List<Peak> peaks) {
        return peaksToMatrix(peaks, false);
    }

    private DoubleMatrix peaksToMatrix(List<Peak> peaks, boolean keepBoundaries) {
        double[][] peakArray = new double[peaks.size()][];
        for (int i = 0; i < peaks.size(); ++i) {
            Peak peak = peaks.get(i);
            peakArray[i] = new double[peak.chromatogram.length];
            int startIndex = 0;
            int endIndex = peak.chromatogram.length - 1;
            if (keepBoundaries) {
                startIndex = Math.max(peak.startIndex, startIndex);
                endIndex = Math.min(peak.endIndex, endIndex);
            }
            for (int j = startIndex; j <= endIndex; ++j) {
                peakArray[i][j] = peak.chromatogram.getIntensity(j);
            }
        }
        return new DoubleMatrix(peakArray).transpose();
    }

    private DoubleMatrix addColumn(DoubleMatrix m, double value) {
        DoubleMatrix newMatrix = DoubleMatrix
                .ones(m.rows, m.columns + 1)
                .muli(value);
        return copyMatrix(m, newMatrix);
    }

    private DoubleMatrix addRow(DoubleMatrix m, double value) {
        DoubleMatrix newMatrix = DoubleMatrix
                .ones(m.rows + 1, m.columns)
                .muli(value);
        return copyMatrix(m, newMatrix);
    }

    private DoubleMatrix copyMatrix(DoubleMatrix m1, DoubleMatrix m2) {
        int numRows = Math.min(m1.rows, m2.rows);
        int numColumns = Math.min(m1.columns, m2.columns);
        for (int i = 0; i < numRows; ++i)
            for (int j = 0; j < numColumns; ++j)
                m2.put(i, j, m1.get(i, j));
        return m2;
    }

    private DoubleMatrix getCoefficientMaxs(DoubleMatrix peakMatrix, DoubleMatrix componentMatrix) {

        int[] retTimeIndices = componentMatrix.columnArgmaxs();

        DoubleMatrix coefficientMaxs = new DoubleMatrix(componentMatrix.columns, peakMatrix.columns);
        for (int i = 0; i < componentMatrix.columns; ++i) {
            for (int j = 0; j < peakMatrix.columns; ++j) {
                coefficientMaxs.put(i, j, peakMatrix.get(retTimeIndices[i], j));
            }
        }

        return coefficientMaxs;
    }

    private DoubleMatrix biGaussiansToMatrix(List<BiGaussian> biGaussians, double[] retTimes) {
        return new DoubleMatrix(
                biGaussians.stream()
                        .map(b -> BiGaussian.evaluate(retTimes, b))
                        .toArray(double[][]::new))
                .transpose();
    }

    private void scaleRows(DoubleMatrix m) {

        for (int i = 0; i < m.rows; ++i) {

            int maxIndex = getRowArgMax(m, i);
            double maxValue = m.get(i, maxIndex);

            if (maxValue <= 0.0) continue;

            for (int j = 0; j < m.columns; ++j) {
                m.put(i, j, m.get(i, j) / maxValue);
            }
        }
    }

    private void flattenLastRow(DoubleMatrix m) {
        int i = m.rows - 1;
        for (int j = 0; j < m.columns; ++j)
            m.put(i, j, 1.0);
    }

    private void truncateColumns(@Nonnull DoubleMatrix m, double threshold) {
        int[] maxIndices = m.columnArgmaxs();
        for (int j = 0; j < m.columns; ++j) {
            double minIntensity = threshold * m.get(maxIndices[j], j);

            int topBound = maxIndices[j] - 1;
            while (topBound >= 0 && m.get(topBound, j) > minIntensity) --topBound;
            for (int i = 0; i < topBound; ++i)
                m.put(i, j, 0.0);

            int bottomBound = maxIndices[j] + 1;
            while (bottomBound < m.rows && m.get(bottomBound, j) > minIntensity) ++bottomBound;
            for (int i = bottomBound; i < m.rows; ++i)
                m.put(i, j, 0.0);
        }
    }

    private void truncateRows(@Nonnull DoubleMatrix m, double threshold) {

        for (int i = 0; i < m.rows; ++i) {

            int maxIndex = getRowArgMax(m, i);

            double minIntensity = threshold * m.get(i, maxIndex);

            int leftBound = maxIndex - 1;
            while (leftBound >= 0 && m.get(i, leftBound) > minIntensity) --leftBound;
            for (int j = 0; j < leftBound; ++j)
                m.put(i, j, 0.0);

            int rightBound = maxIndex + 1;
            while (rightBound < m.columns && m.get(i, rightBound) > minIntensity) ++rightBound;
            for (int j = rightBound; j < m.columns; ++j)
                m.put(i, j, 0.0);
        }
    }

    private void makeUnimodalColumns(DoubleMatrix m) {
        int[] maxIndices = m.columnArgmaxs();
        for (int j = 0; j < m.columns; ++j) {

            for (int i = maxIndices[j] - 1; i >= 0; --i)
                m.put(i, j, m.get(i, j) < m.get(i + 1, j)
                        ? m.get(i, j)
                        : m.get(i + 1, j));

            for (int i = maxIndices[j] + 1; i < m.rows; ++i)
                m.put(i, j, m.get(i, j) < m.get(i - 1, j)
                        ? m.get(i, j)
                        : m.get(i - 1, j));
        }
    }

    private void makeUnimodalRows(DoubleMatrix m) {

        for (int i = 0; i < m.rows; ++i) {

            int maxIndex = getRowArgMax(m, i);

            for (int j = maxIndex - 1; j >= 0; --j)
                m.put(i, j, m.get(i, j) < m.get(i, j + 1)
                        ? m.get(i, j)
                        : m.get(i, j + 1));

            for (int j = maxIndex + 1; j < m.columns; ++j)
                m.put(i, j, m.get(i, j) < m.get(i, j - 1)
                        ? m.get(i, j)
                        : m.get(i, j - 1));
        }
    }

    private int getRowArgMax(DoubleMatrix m, int row) {

        double maxValue = -Double.MAX_VALUE;
        int maxIndex = -1;
        for (int j = 0; j < m.columns; ++j)
            if (m.get(row, j) > maxValue) {
                maxValue = m.get(row, j);
                maxIndex = j;
            }

        if (maxIndex < 0)
            throw new IllegalStateException("Cannot find maximum value");

        return maxIndex;
    }

    private void smoothRows(DoubleMatrix m) {
        DoubleMatrix mCopy = m.dup();
        final int shift = SMOOTHING_COEFFICIENTS.length / 2;
        for (int i = 0; i < m.rows; ++i)
            for (int j = 2; j < m.columns - 2; ++j) {
                double v = 0.0;
                for (int k = 0; k < SMOOTHING_COEFFICIENTS.length; ++k)
                    v += mCopy.get(i, j - shift + k) * SMOOTHING_COEFFICIENTS[k];
                m.put(i, j, v > 0.0 ? v : 0.0);
            }
    }


    private double[] getRetTimes(List<BetterPeak> peaks, double startRetTime, double endRetTime) {

        Set<Double> retTimeSet = new HashSet<>();
        for (BetterPeak peak : peaks)
            for (double retTime : peak.chromatogram.xs)
                if (startRetTime <= retTime && retTime <= endRetTime)
                    retTimeSet.add(retTime);

        List<Double> retTimeList = new ArrayList<>(retTimeSet);
        retTimeList.sort(Comparator.naturalOrder());

        return retTimeList.stream().mapToDouble(Double::doubleValue).toArray();
    }
}
